import pennylane as qml
import numpy as np

def print_circuit(circuit):
    weights = np.random.random(size=circuit.weight_shape['weights'])
    inputs  = np.random.random(size=(circuit.n_qubits))
    print(qml.draw(circuit.circuit, expansion_strategy="device")(inputs, weights))




import pennylane as qml

def ai_interface(circuit, weight_shape, n_qubits, framework):
    
    if framework == 'torch':
        qlayer = qml.qnn.TorchLayer(circuit, weight_shape)
    elif framework == 'keras':
        qlayer = qml.qnn.KerasLayer(circuit, weight_shape, output_dim=n_qubits)
    else:
        raise Exception('Framerwork can be only torch or keras!')
    return qlayer



import pennylane as qml
import torch

import sys
sys.path += ['.', './layers/']


class BasicMLP(torch.nn.Module):
    def __init__(self, circuit):
        super().__init__()

        n_qubits = circuit.n_qubits
        self.fc_1 = torch.nn.Linear(n_qubits, n_qubits)
        self.qc_1 = circuit.qlayer
        self.fc_2 = torch.nn.Linear(n_qubits, n_qubits)
        self.softmax  = torch.nn.Softmax(dim=1)

    def forward(self, x):
        x = self.fc_1(x)
        x = self.qc_1(x)
        x = self.fc_2(x)
        return self.softmax(x)
    



import os, pathlib
from setuptools.command.develop import develop
from setuptools.command.install import install
from setuptools.command.sdist import sdist
from setuptools import setup, find_packages
from subprocess import check_call

def readme():
    """ Generate readme file. """
    try:
        with open("./README.md", encoding="utf8") as file:
            return file.read()
    except IOError:
        return ""

# Package 
HERE = pathlib.Path(__file__).parent
PACKAGE_NAME  = 'hqm'
VERSION = '0.0.2'
AUTHOR = 'Alessandro Sebastianelli'
AUTHOR_EMAIL = 'alessandro.sebastianelli1995@gmail.com'
URL = 'https://github.com/alessandrosebastianelli/hybrid_quantum_models.git'

LICENSE = 'MIT'
DESCRIPTION = 'Hybrid Quantum Models - HQM'
LONG_DESCRIPTION = readme()
LONG_DESC_TYPE = 'text/markdown'


INSTALL_REQUIRES = [
	'numpy',
	'pennylane',
]


def gitcmd_update_submodules():
	'''	Check if the package is being deployed as a git repository. If so, recursively
		update all dependencies.

		@returns True if the package is a git repository and the modules were updated.
			False otherwise.
	'''
	if os.path.exists(os.path.join(HERE, '.git')):
		check_call(['git', 'submodule', 'update', '--init', '--recursive'])
		return True

	return False


class gitcmd_develop(develop):
	'''	Specialized packaging class that runs git submodule update --init --recursive
		as part of the update/install procedure.
	'''
	def run(self):
		gitcmd_update_submodules()
		develop.run(self)


class gitcmd_install(install):
	'''	Specialized packaging class that runs git submodule update --init --recursive
		as part of the update/install procedure.
	'''
	def run(self):
		gitcmd_update_submodules()
		install.run(self)


class gitcmd_sdist(sdist):
	'''	Specialized packaging class that runs git submodule update --init --recursive
		as part of the update/install procedure;.
	'''
	def run(self):
		gitcmd_update_submodules()
		sdist.run(self)


setup(
    cmdclass={
		'develop': gitcmd_develop, 
		'install': gitcmd_install, 
		'sdist': gitcmd_sdist,
	}, 
	name=PACKAGE_NAME,
	version=VERSION,
	description=DESCRIPTION, 
	long_description_content_type=LONG_DESC_TYPE,
	author=AUTHOR, license=LICENSE, 
	author_email=AUTHOR_EMAIL, 
	url=URL, 
	install_requires=INSTALL_REQUIRES,
	packages=find_packages())


from types import FunctionType
import pennylane as qml
import numpy as np
import warnings
import sys

sys.path += ['.', './layers/']

from hqm.layers.ai_interface import ai_interface

class BasicEntangledCircuit:
    '''
        This class implements a torch/keras quantum layer using a basic entangler
        circuit. 
    '''
    
    def __init__(self, n_qubits : int, n_layers : int, aiframework : str, dev : qml.device = None) -> None:
        '''
            BasicEntangledCircuit constructor.  

            Parameters:  
            -----------
            - n_qubits : int  
                number of qubits for the quantum circuit  
            - n_layers : int  
                number of layers for the quantum circuit  
            - aiframework : str  
                string representing the AI framework in use, can be 'torch' or 'keras'. This will create  
                a compatible trainable layer for the framework.   
            - dev : qml.device  
                PennyLane device on wich run quantum operations (dafault : None). When None it will be set
                to 'default.qubit'  
            
            Returns:  
            --------  
            Nothing, a BasicEntangledCircuit object will be created.  
        '''

        # Checking for exceptions  
        if aiframework not in ['torch', 'keras']: raise Exception(f"Accepted values for framerwork are 'torch' or 'keras', found {aiframework}")
        if n_qubits < 1: raise Exception(f"Number of qubits must be greater or equal than 1, found {n_qubits}")
        if n_layers < 1: raise Exception(f"Number of layers must be greater or equal than 1, found {n_layers}")

        # Set dev to 'default.qubit' if dev is None  
        if dev is None: 
            dev = qml.device("default.qubit", wires=n_qubits)
            warnings.warn(f"Dev has been set to None, setting it to {dev}")

        self.n_qubits     = n_qubits
        self.n_layers     = n_layers
        self.aiframework  = aiframework
        self.dev          = dev
        self.weight_shape = {"weights": (n_layers, n_qubits)}
        self.circuit      = self.circ(self.dev, self.n_qubits)
        self.qlayer       = ai_interface(circuit      = self.circuit, 
                                         weight_shape = self.weight_shape, 
                                         n_qubits     = self.n_qubits, 
                                         framework    = self.aiframework)

    @staticmethod
    def circ(dev : qml.device, n_qubits : int) -> FunctionType:
        '''
            BasicEntangledCircuit static method that implements the quantum circuit.  

            Parameters:  
            -----------  
            - dev : qml.device  
                PennyLane device on wich run quantum operations (dafault : None). When None it will be set  
                to 'default.qubit'  
            - n_qubits : int  
                number of qubits for the quantum circuit  

            Returns:  
            --------  
            - qnode : qml.qnode  
                the actual PennyLane circuit   
        '''
        @qml.qnode(dev)
        def qnode(inputs : np.ndarray, weights : np.ndarray) -> list:
            '''
                PennyLane based quantum circuit composed of an angle embedding layer and a basic entangler
                layer.  

                Parameters:  
                -----------  
                - inputs : np.ndarray  
                    array containing input values (can came from previous torch/keras layers or quantum layers)  
                - weights : np.ndarray  
                    array containing the weights of the circuit that are tuned during training, the shape of this
                    array depends on circuit's layers and qubits.   
                
                Returns:  
                --------  
                - measurements : list  
                    list of values measured from the quantum circuits  
            '''
            qml.AngleEmbedding(inputs, wires=range(n_qubits))
            qml.BasicEntanglerLayers(weights, wires=range(n_qubits))
            measurements = [qml.expval(qml.PauliZ(wires=i)) for i in range(n_qubits)]
            return measurements
    
        return qnode
    
class StronglyEntangledCircuit:
    '''
        This class implements a torch/keras quantum layer using a strongly entangler
        circuit.
    '''

    def __init__(self, n_qubits : int, n_layers : int, aiframework : str, dev : qml.device = None) -> None:
        '''
            StronglyEntangledCircuit constructor.  

            Parameters:  
            -----------  
            - n_qubits : int  
                number of qubits for the quantum circuit  
            - n_layers : int  
                number of layers for the quantum circuit  
            - aiframework : str  
                string representing the AI framework in use, can be 'torch' or 'keras'. This will create
                a compatible trainable layer for the framework.  
            - dev : qml.device  
                PennyLane device on wich run quantum operations (dafault : None). When None it will be set
                to 'default.qubit'  
            
            Returns:  
            --------  
            Nothing, a StronglyEntangledCircuit object will be created.  
        '''

        # Checking for exceptions  
        if aiframework not in ['torch', 'keras']: raise Exception(f"Accepted values for framerwork are 'torch' or 'keras', found {aiframework}")
        if n_qubits < 1: raise Exception(f"Number of qubits must be greater or equal than 1, found {n_qubits}")
        if n_layers < 1: raise Exception(f"Number of layers must be greater or equal than 1, found {n_layers}")

        # Set dev to 'default.qubit' if dev is None  
        if dev is None: 
            dev = qml.device("default.qubit", wires=n_qubits)
            warnings.warn(f"Dev has been set to None, setting it to {dev}")

        self.n_qubits     = n_qubits
        self.n_layers     = n_layers
        self.aiframework  = aiframework
        self.dev          = dev
        self.weight_shape = {"weights": (n_layers, n_qubits, 3)}
        self.circuit      = self.circ(self.dev, self.n_qubits)
        self.qlayer       = ai_interface(circuit      = self.circuit, 
                                         weight_shape = self.weight_shape, 
                                         n_qubits     = self.n_qubits, 
                                         framework    = self.aiframework)
        
    @staticmethod
    def circ(dev : qml.device, n_qubits : int) -> FunctionType:
        '''
            StronglyEntangledCircuit static method that implements the quantum circuit.  

            Parameters:  
            -----------  
            - dev : qml.device  
                PennyLane device on wich run quantum operations (dafault : None). When None it will be set
                to 'default.qubit'  
            - n_qubits : int  
                number of qubits for the quantum circuit  

            Returns:  
            --------  
            - qnode : qml.qnode  
                the actual PennyLane circuit  
        '''
        @qml.qnode(dev)
        def qnode(inputs : np.ndarray, weights : np.ndarray) -> list:
            '''
                PennyLane based quantum circuit composed of an angle embedding layer and a strongly entangler
                layer.  

                Parameters:  
                -----------  
                - inputs : np.ndarray  
                    array containing input values (can came from previous torch/keras layers or quantum layers)  
                - weights : np.ndarray  
                    array containing the weights of the circuit that are tuned during training, the shape of this
                    array depends on circuit's layers and qubits.   
                
                Returns:  
                --------  
                - measurements : list   
                    list of values measured from the quantum circuits  
            '''
            qml.AngleEmbedding(inputs, wires=range(n_qubits))
            qml.StronglyEntanglingLayers(weights, wires=range(n_qubits))
            measurements = [qml.expval(qml.PauliZ(wires=i)) for i in range(n_qubits)]
            return measurements
    
        return qnode
